# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# Written (W) 2006-2007 Soeren Sonnenburg
# Written (W) 2007 Gunnar Raetsch
# Copyright (C) 2007 Fraunhofer Institute FIRST and Max-Planck-Society
# 

import sys
import numpy
import seqdict
import model

from shogun.Classifier import SVM
from shogun.Features import CharFeatures,DNA
from shogun.Kernel import WeightedDegreeCharKernel

class svm_splice_model(object):
	def __init__(self, order, traindat, alphas, b, (window_left,offset,window_right), consensus):

		f=CharFeatures(traindat, DNA)
		wd_kernel = WeightedDegreeCharKernel(f,f, int(order))
		wd_kernel.io.set_target_to_stderr()

		self.svm=SVM(wd_kernel, alphas, numpy.arange(len(alphas), dtype=numpy.int32), b)
		self.svm.io.set_target_to_stderr()
		self.svm.parallel.set_num_threads(self.svm.parallel.get_num_cpus())
		#self.svm.set_linadd_enabled(False)
		#self.svm.set_batch_computation_enabled(true)

		self.window_left=int(window_left)
		self.window_right=int(window_right)

		self.consensus=consensus
		self.wd_kernel=wd_kernel
		self.traindat=f
		self.offset = offset ;

	def get_positions(self, sequence):
		positions=list()

		for cons in self.consensus:
			l=sequence.find(cons)
			while l>-1:
				#if l<len(sequence)-self.window_right-2 and l>self.window_left:
				positions.append(l+self.offset)
				l=sequence.find(cons, l+1)

		positions.sort()
		return positions

	def get_predictions_from_seqdict(self, seqdic, site):
		""" we need to generate a huge test features object 
			containing all locations found in each seqdict-sequence
			and each location (this is necessary to efficiently
			(==fast,low memory) compute the splice outputs
		"""

		seqlen=self.window_right+self.window_left+2

		num=0
		for s in seqdic:
			num+= len(s.preds[site].positions)

		testdat = numpy.chararray((seqlen,num), 1, order='FORTRAN')

		k=0
		si = 0 ;
		for s in seqdic:
			sequence=s.seq
			positions=s.preds[site].positions
			si += 1 
			for j in xrange(len(positions)):
				if len(positions)>50000 and j%10000==0:
					sys.stderr.write('sequence %i: progress %1.2f%%\r' %(si,(100*j/len(positions))))
				i=positions[j] - self.offset
				start = i-self.window_left
				if start<0:
					s_start='A'*(-start)
					start =  0;
				else:
					s_start = ''
				stop = i+self.window_right+2
				if stop>len(sequence):
					s_stop = 'A'*(stop-len(sequence) +1)
					stop=len(sequence) - 1 ;
				else:
					s_stop = '' ;
				s= s_start + sequence[start:stop] + s_stop
				#print len(s)
				#s=sequence[i-self.window_left:i+self.window_right+2]
				testdat[:,k]=list(s)
				k+=1
				
		if len(positions)>50000:
			sys.stderr.write('\n')

		t=CharFeatures(testdat, DNA)

		self.wd_kernel.init(self.traindat, t)
		l=self.svm.classify().get_labels()
		sys.stderr.write("\n...done...\n")

		k=0
		for s in seqdic:
			num=len(s.preds[site].positions)
			scores= num * [0]
			for j in xrange(num):
				scores[j]=l[k]
				k+=1
			s.preds[site].set_scores(scores)

	def get_positions_from_seqdict(self, seqdic, site):
		for d in seqdic:
			positions=list()
			sequence=d.seq
			for cons in self.consensus:
				l=sequence.find(cons)
				while l>-1:
					#if l<len(sequence)-self.window_right-2 and l>self.window_left:
					positions.append(l+self.offset)
					l=sequence.find(cons, l+1)
			positions.sort()
			d.preds[site].set_positions(positions)

	def get_predictions(self, sequence, positions):

		seqlen=self.window_right+self.window_left+2
		num=len(positions)

		testdat = numpy.chararray((seqlen,num),1,order='FORTRAN')

		for j in xrange(num):
			i=positions[j] - self.offset ;
			start = i-self.window_left
			if start<0:
				s_start='A'*(-start)
				start =  0;
			else:
				s_start = ''
			stop = i+self.window_right+2
			if stop>len(sequence):
				s_stop = 'A'*(stop-len(sequence))
				stop=len(sequence) - 1 ;
			else:
				s_stop = '' ;
			s= s_start + sequence[start:stop] + s_stop
			print len(s)
			testdat[:,j]=list(s)
		t=CharFeatures(testdat, DNA)

		self.wd_kernel.init(self.traindat, t)
		l=self.svm.classify().get_labels()
		sys.stderr.write("\n...done...\n")
		return l

class signal_detectors(object):
	def __init__(self, model, donor_splice_use_gc):
		if donor_splice_use_gc:
			donor_consensus=['GC','GT']
		else:
			donor_consensus=['GT']
		
		self.acceptor=svm_splice_model(model.acceptor_splice_order, model.acceptor_splice_svs,
				numpy.array(model.acceptor_splice_alphas).flatten(), model.acceptor_splice_b, 
				(model.acceptor_splice_window_left, 2, model.acceptor_splice_window_right), ['AG'])
		self.donor=svm_splice_model(model.donor_splice_order, model.donor_splice_svs, 
				numpy.array(model.donor_splice_alphas).flatten(), model.donor_splice_b, 
				(model.donor_splice_window_left, 0, model.donor_splice_window_right),
				donor_consensus)

	def set_sequence(self, seq):
		self.acceptor.set_sequence(seq)
		self.donor.set_sequence(seq)

	def predict_acceptor_sites(self, seq):
		pos=self.acceptor.get_positions(seq)
		sys.stderr.write("computing svm output for acceptor positions\n")
		pred=self.acceptor.get_predictions(seq, pos)
		return (pos,pred)

	def predict_donor_sites(self,seq):
		pos=self.donor.get_positions(seq)
		sys.stderr.write("computing svm output for donor positions\n")
		pred=self.donor.get_predictions(seq, pos)
		return (pos,pred)

	def predict_acceptor_sites_from_seqdict(self, seqs):
		self.acceptor.get_positions_from_seqdict(seqs, 'acceptor')
		sys.stderr.write("computing svm output for acceptor positions\n")
		self.acceptor.get_predictions_from_seqdict(seqs, 'acceptor')

	def predict_donor_sites_from_seqdict(self, seqs): 
		self.donor.get_positions_from_seqdict(seqs, 'donor')
		sys.stderr.write("computing svm output for donor positions\n")
		self.donor.get_predictions_from_seqdict(seqs, 'donor')

