#!/usr/bin/env python2.4
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# Written (W) 2006-2007 Uta Schulze, Cheng Soon Ong, Gunnar Raetsch 
# Copyright (C) 2006 Max Planck Society
# 

#<------------------------------------------------------------------------------>
# Open Questions and TODO
#
# precompute memory usage and give warnings
#<------------------------------------------------------------------------------>

import pdb

import sys
import bz2
import gzip
import os
import math
import string
import time
import cPickle
import array
from copy import copy
from optparse import OptionParser
import subprocess

import palma
import palma.genomic
import palma.signal_detectors
import palma.seqdict
import palma.model
import palma.output_formating
from palma.signal_detectors import signal_detectors

try:
    import numpy
except:
    sys.stderr.write("'import numpy' failed\n")
    sys.exit(2)


#<------------------------------------------------------------------------------>
# CONSTANTS
#<------------------------------------------------------------------------------>
#nr_paths = 1                                    #number of alignments computed
param_filename_ending = '.param.bz2'
translation = '-ACGTN_'                          #how aligned est and dna sequence is displayed
                                                 #(gap_char, 5 nucleotides, intron_char)  
comparison_char = '| '                           #first: match_char, second: intron_char
#<------------------------------------------------------------------------------>


#<------------------------------------------------------------------------------>
# organism dependent constants
#<------------------------------------------------------------------------------>
organism_constants = {'elegans': {'max_gene_length' : 50000,\
								  'version'         : 'WS150',},\
					  'elegans_micro': {'max_gene_length' : 50000,\
								  'version'         : 'WS150',},\
					  'human':   {'max_gene_length' : 100000,\
								  'version'         : 'hg17',}}
#<------------------------------------------------------------------------------>

class palma_align:
    def __init__(self):
        self.model = None
        self.signal = None
        self.donor_splice_use_gc = None
                
    def load_model(self, filenames, parameters):
        #<------------------------------------------------------------------------------>
        # loading model 
        #<------------------------------------------------------------------------------>
        try:
            param_file = file(filenames['data_dir']+filenames['param'])
        except IOError:
            sys.stderr.write("Can't open PARAM FILE (%s) to read\n" %(filenames['data_dir']+filenames['param']))
            sys.exit(2)

        #<------------------------------------------------------------------------------>
        filename = filenames['data_dir']+filenames['param'] #'palma_elegansWS120_gc=1_orf=0.dat' ;
        #filename = 'test.dat.bz2' 
        f=None
        picklefile=filename+'.pickle'
        if os.path.isfile(picklefile):
            self.model=cPickle.load(file(picklefile))       
        else:
            if filename.endswith('.bz2'):
                f=bz2.BZ2File(filename);
            else:
                f=file(filename);
                                
            self.model = palma.model.parse_file(f)
            f.close()
            f=file(picklefile,'w')
            cPickle.dump(self.model, f)
            f.close()
                        
        self.donor_splice_use_gc = parameters['use_gc'] ;
        #self.plif=plif(self.model)
        if parameters['use_splicesites']:
            self.signal = signal_detectors(self.model, self.donor_splice_use_gc)

    def xdformat(self, filenames, parameters):
        if len(parameters['blast_dbfile']) == 0:
            sys.stdout.write("Creating databases for BLAST in XDF (eXtended Database Format)\n")
            all_chrom_files = filenames['dna']
            if len(parameters['blast_dbfile'])==0:
                parameters['blast_dbfile'] = filenames['dna']
            #all_chrom_files = reduce(lambda x,y: x+y, map(lambda x: ' ' + x + ' ', filenames['dna']))
            xdformat_command = [parameters['xdformat'], ' -n -o ' + parameters['blast_dbfile'] + ' -I ' + all_chrom_files] 
            try:
                print xdformat_command[0], xdformat_command[1]
                # subprocess.call doesn't seem to work properly on OSX
                #subprocess.call(xdformat_command)
                os.popen(xdformat_command[0]+xdformat_command[1])
            except:
                sys.stderr.write("ERROR: '%s %s' failed\n" %(xdformat_command[0],xdformat_command[1]))
                sys.exit(2)

    def blastn(self, filenames, parameters):
        #<------------------------------------------------------------------------------>
        # blastn
        #<------------------------------------------------------------------------------>
        blastn_command = [parameters['blastn'], ' ' + parameters['blast_dbfile'] + ' ' + \
                          filenames['est'] + ' -mformat=2,' + filenames['blast_out'] + \
                          ' -sort_by_highscore -hspsepSmax=' + str(parameters['max_gene_length'])\
                          + ' -topcomboN=' + str(parameters['topcomboN']) + ' -notes -warnings']

        try:
            print blastn_command[0], blastn_command[1]
            # subprocess.call doesn't seem to work properly on OSX
            #subprocess.call(blastn_command)
            os.popen(blastn_command[0] + blastn_command[1])
        except:
            sys.stderr.write("ERROR: '%s %s' failed\n" %(blastn_command[0],blastn_command[1]))
            sys.exit(2)       

        #<------------------------------------------------------------------------------>
        # blastn output
        #<------------------------------------------------------------------------------>
        try:
            blast_out_file = open(filenames['blast_out'],'r')
            output = blast_out_file.readlines()
            blast_out_file.close()
        except IOError:
            sys.stderr.write("Can't read  %s \n" %filenames['blast_out'])
            sys.exit(2)
                        
        blast_result = palma.palma_utils.parse_wublast(output, parameters['max_e_value'])

        return blast_result


    def save_splicescores(self, filename, splicescores):
        index={} ;
        
        if filename.endswith('.bz2'):
            f=bz2.BZ2File(filename, 'wb+');
        else:
            if filename.endswith('.gz'):
                f=gzip.GzipFile(filename, 'wb+');
            else:
                f = file(filename, 'wb+')

        posarray = array.array('i') ;
        posarray.fromlist([0])
        posarray.tofile(f)
                
        for key in splicescores.keys():
            #print key
            start = f.tell()
            a = array.array('f')
            a.fromlist(splicescores[key]) 
            a.tofile(f)
            end = f.tell()
            index[key] = (len(splicescores[key]),start,end)

        # dump index file
        pos = f.tell() ;
        cPickle.dump(index, f)
        f.seek(0,0) ; # overwrite the index position
        posarray = array.array('i') ;
        posarray.fromlist([pos])
        posarray.tofile(f)
        f.close

    def load_splicescores(self, filename):
        if filename.endswith('.bz2'):
            f=bz2.BZ2File(filename, 'rb');
        else:
            if filename.endswith('.gz'):
                f=gzip.GzipFile(filename, 'rb');
            else:
                f = file(filename, 'rb')

        posarray = array.array('i') ;
        posarray.read(f, 1)
        pos = posarray[0]
        #print pos
        f.seek(pos, 0)
        index = cPickle.load(f)

        splicescores = {} ;
        for key in index.keys():
            #print key
            (num, start, end) = index[key] 
            f.seek(start, 0)
            a = array.array('f')
            a.read(f, num)
            splicescores[key] = a.tolist()
            assert(f.tell() <= end) 

        f.close()
        return splicescores


    def load_splicescores_index(self, filename):
        if filename.endswith('.bz2'):
            f=bz2.BZ2File(filename, 'rb');
        else:
            if filename.endswith('.gz'):
                f=gzip.GzipFile(filename, 'rb');
            else:
                f = file(filename, 'rb')

        posarray = array.array('i') ;
        posarray.read(f, 1)
        pos = posarray[0]
        #print pos
        f.seek(pos, 0)
        index = cPickle.load(f)
        f.close()
        index['XXX_palma_filename'] = filename 

        return index
                
    def get_splicescores(self, index, key, start, end):
        if not index.has_key('XXX_palma_filename'):
            seq = index[key] ;
            if end==start==0:
                return seq
            else:
                return seq[start:end]
        else:
            (num, fstart, fend) = index[key] ;
            filename = index['XXX_palma_filename']
            if filename.endswith('.bz2'):
                f = bz2.BZ2File(filename, 'rb');
            else:
                if filename.endswith('.gz'):
                    f = gzip.GzipFile(filename, 'rb');
                else:
                    f = file(filename, 'rb')

            f.seek(fstart + start*4, 0) # sizeof float = 4!?
            ss = array.array('f') ;
            if end==start==0:
                ss.read(f, num)
            else:
                ss.read(f, end-start)

            return ss.tolist()

    def predict_splicescores(self, filenames, parameters, dna_fasta_dict):
        #t0 = time.time() ;
        splicescores_picklefile = filenames['dna'] + '.palma_splicesites.sspickle' ;
        #if os.path.isfile(splicescores_picklefile):
        if parameters['use_splicesites'] and os.path.isfile(splicescores_picklefile):
            print 'loading splice site predictions from ' + splicescores_picklefile
            splicescores = self.load_splicescores_index(splicescores_picklefile)
            splicescores_precomputed = True
        else:
            splicescores = {} ;
            splicescores_precomputed = False 
                        
            #get donor/acceptor signal predictions for all sequences
            if parameters['use_splicesites']:
                seqs= palma.seqdict.seqdict(dna_fasta_dict)
                self.signal.predict_acceptor_sites_from_seqdict(seqs)
                self.signal.predict_donor_sites_from_seqdict(seqs)

        #print time.time()-t0
                
        for counter in xrange(len(dna_fasta_dict.keys())):
            #<------------------------------------------------------------------------------>
            # DNA and mRNA sequence
            #<------------------------------------------------------------------------------>       
            if parameters['use_splicesites'] and not splicescores_precomputed:
                seq = seqs[counter]
            tSeq = dna_fasta_dict[dna_fasta_dict.keys()[counter]] # dna_sequences is list of strings
            #assert(seq.sequence == tSeq)
                        
            tName = dna_fasta_dict.keys()[counter] 
            tSize = len(tSeq)
            #print tName
			
            #<------------------------------------------------------------------------------>
            # Splice site scores
            #<------------------------------------------------------------------------------>
            if parameters['use_splicesites']:
                if splicescores_precomputed:
                    assert(splicescores.has_key(tName + '_acceptor_splicescores'))
                    assert(splicescores.has_key(tName + '_donor_splicescores'))
                else:
                    acc_support_values = [-numpy.inf]*tSize
                    for i in xrange(len(seq.preds['acceptor'].positions)):
                        acc_support_values[seq.preds['acceptor'].positions[i]-1] = seq.preds['acceptor'].scores[i] ;
                                                
                    don_support_values = [-numpy.inf]*tSize
                    for i in xrange(len(seq.preds['donor'].positions)):
                        don_support_values[seq.preds['donor'].positions[i]] = seq.preds['donor'].scores[i] ;
            else:
                don_support_values = [-numpy.inf]*tSize
                for cnt in range(tSize-1):
                    if (tSeq[cnt] in 'Gg') and ((tSeq[cnt+1] in 'Tt') or ((tSeq[cnt+1] in 'cC') and parameters['use_gc'])):
                        don_support_values[cnt] = 1
                assert(len(don_support_values)==tSize)

                acc_support_values = [-numpy.inf]*tSize
                for cnt in range(tSize-1):
                    if (tSeq[cnt+1] in 'Gg') and (tSeq[cnt] in 'aA'):
                        acc_support_values[cnt+1] = 1
                assert(len(acc_support_values)==tSize)

            if not splicescores.has_key(tName + '_acceptor_splicescores'):
                splicescores[tName + '_acceptor_splicescores'] = acc_support_values 
                                
            if not splicescores.has_key(tName + '_donor_splicescores'):
                splicescores[tName + '_donor_splicescores'] = don_support_values 


            #print time.time()-t0

        # write the splice site predictions into a pickle file
        #if (not splicescores_precomputed) and parameters['save_splicescores']:
        if (not splicescores_precomputed) and (parameters['use_splicesites']) and parameters['save_splicescores']:
            print 'saving splice site scores to ' + splicescores_picklefile
            self.save_splicescores(splicescores_picklefile, splicescores)

        #print time.time()-t0
		
        return splicescores


    def predict_file_paired(self, filenames, parameters, out_file, bed_file, dna_fasta_dict, est_fasta_dict, splicescores):
        if len(est_fasta_dict) != len(dna_fasta_dict):
            sys.stderr.write("ERROR: Number of entries in %s (%d) and %s (%d) different\n"\
                             %(filenames['dna'], len(dna_fasta_dict), filenames['est'], len(est_fasta_dict)))
            sys.exit(2)

        #<------------------------------------------------------------------------------>
        # Output on stdout
        #<------------------------------------------------------------------------------>
        #sys.stdout.write("Loaded %i letters in %i sequences\n" %(numpy.sum(map(lambda x: len(x), dna_sequences)), dna_entries))
        #sys.stdout.write("Searching %i bases in %i sequences\n" %(numpy.sum(map(lambda x: len(x), est_sequences)), est_entries))
        #<------------------------------------------------------------------------------>

        #if parameters['print_header'] and (not parameters['print_sequences']):
        #    palma.output_formating.print_header(out_file) #the very first header and the only one, if print_sequences==0



        for counter in xrange(len(dna_fasta_dict.keys())):
            #<------------------------------------------------------------------------------>
            # DNA and mRNA sequence
            #<------------------------------------------------------------------------------>       
            tSeq = dna_fasta_dict[dna_fasta_dict.keys()[counter]] # dna_sequences is list of strings
            tName = dna_fasta_dict.keys()[counter] 
            tSize = len(tSeq)
            
            qSeq = est_fasta_dict[est_fasta_dict.keys()[counter]] # est_sequences is list of strings
            qName = est_fasta_dict.keys()[counter]
            qSize = len(qSeq)

            #<------------------------------------------------------------------------------>
            # Splice site scores
            #<------------------------------------------------------------------------------>
            #acc_support_values = splicescores[tName + '_acceptor_splicescores'] 
            #don_support_values = splicescores[tName + '_donor_splicescores'] 
            acc_support_values = self.get_splicescores(splicescores, tName + '_acceptor_splicescores', 0, 0)
            don_support_values = self.get_splicescores(splicescores, tName + '_donor_splicescores', 0, 0)
            #<------------------------------------------------------------------------------>
                        
            #<------------------------------------------------------------------------------>
            # get alignment information for nr_path many alignments
            #<------------------------------------------------------------------------------>       
            align_info = palma.alignment_info()
            try:
                reduce_region = (tSize*qSize*8 > parameters['reduce_threshold'])
                align_info = palma.do_alignment(tSeq, tSize, qSeq, qSize, reduce_region,\
                                                self.model, '+', translation, comparison_char,\
                                                don_support_values, acc_support_values)

            except MemoryError:
                sys.stderr.write("WARNING: Not enough memory for creating the alignment matrices\n")
                # if dna_entries > 1:
                #       sys.stderr.write("SKIPPING entry %d\n" %(counter+1))
                #       continue
                sys.exit(2)

            for ai in align_info: # output all paths
                palma.output_formating.print_results(ai, out_file, bed_file, tName, tSize, \
                                                     qName, qSize, '+', '+', parameters)
                                

    def predict_file_all(self, filenames, parameters, out_file, bed_file, dna_fasta_dict, est_fasta_dict, splicescores):
        #<------------------------------------------------------------------------------>
        # Output on stdout
        #<------------------------------------------------------------------------------>
        #sys.stdout.write("Loaded %i letters in %i sequences\n" %(numpy.sum(map(lambda x: len(x), dna_sequences)), dna_entries))
        #sys.stdout.write("Searching %i bases in %i sequences\n" %(numpy.sum(map(lambda x: len(x), est_sequences)), est_entries))
        #<------------------------------------------------------------------------------>

        #if parameters['print_header'] and (not parameters['print_sequences']):
        #    palma.output_formating.print_header(out_file) #the very first header and the only one, if print_sequences==0

        total_alignments = len(est_fasta_dict.keys())*len(dna_fasta_dict.keys())
        if not parameters['respectQueryOrientation']: total_alignments *= 2
        num_alignments = 0
		
        for est_counter in xrange(len(est_fasta_dict.keys())):
            best_score = -1e10 
            best_tName = None
            best_tSize = None 
            best_alignment = None
            best_tStrand = None 
            best_qStrand = None 
                        
            #<------------------------------------------------------------------------------>
            # DNA and mRNA sequence
            #<------------------------------------------------------------------------------>       
            qSeq = est_fasta_dict[est_fasta_dict.keys()[est_counter]] # est_sequences is list of strings
            qName = est_fasta_dict.keys()[est_counter]
            qSize = len(qSeq)
            #print qName
            for qStrand in '+-':

                if (parameters['respectQueryOrientation']==True) and (qStrand == '-'):
                    continue 
                if qStrand == '-':
                    qSeq = palma.palma_utils.reverse_complement(qSeq) 

                for dna_counter in xrange(len(dna_fasta_dict.keys())):
                    tSeq = dna_fasta_dict[dna_fasta_dict.keys()[dna_counter]] # dna_sequences is list of strings
                    tName = dna_fasta_dict.keys()[dna_counter]
                    tSize = len(tSeq)
                    #print tName

                    #<------------------------------------------------------------------------------>
                    # Splice site scores
                    #<------------------------------------------------------------------------------>
                    acc_support_values = self.get_splicescores(splicescores, tName + '_acceptor_splicescores', 0, 0)
                    don_support_values = self.get_splicescores(splicescores, tName + '_donor_splicescores', 0, 0)
                    #<------------------------------------------------------------------------------>

                    tStrand = '+' ;
                    if tName.endswith('_reverse'):
                        tStrand = '-'
                        tName = tName[0:-8]

                    #<------------------------------------------------------------------------------>
                    # get alignment information for nr_path many alignments
                    #<------------------------------------------------------------------------------>
                    num_alignments += 1
                    print 'Performing alignment', num_alignments, 'of', total_alignments, ':',\
                          qName, qStrand, 'vs.', tName, tStrand
                    try:
                        reduce_region = True; #(tSize*qSize*8 > parameters['reduce_threshold'])
                        align_info = palma.do_alignment(tSeq, tSize, qSeq, qSize, reduce_region,\
                                                        self.model, '+', translation, comparison_char,\
                                                        don_support_values, acc_support_values)
                        if (len(align_info)>0):
                            #print align_info[0].score
                            if align_info[0].score>best_score:
                                best_score = align_info[0].score
                                best_tName = tName
                                best_tSize = tSize 
                                best_alignment = align_info[0]
                                best_qStrand = qStrand 
                                best_tStrand = tStrand 

                    except MemoryError:
                        sys.stderr.write("WARNING: Not enough memory for creating the alignment matrices\n")
                        sys.exit(2)

            # output best match
            if best_alignment is not None:
				palma.output_formating.print_results(best_alignment, out_file, bed_file, \
                                    best_tName, best_tSize, qName, qSize, best_qStrand, best_tStrand, parameters)


    def predict_file_blast(self, filenames, parameters, out_file, bed_file, dna_fasta_dict, est_fasta_dict, splicescores, blast_result):
        if len(blast_result.q_name)==0:
            sys.stderr.write('no blast hit\n')
            return
		
        q_names = copy(blast_result.q_name)
        q_name = q_names.pop(0)
        q_name_list = [q_name] ;
        while (len(q_names) > 0):
            q_name_ = q_names.pop(0)
            if not q_name == q_name_:
                q_name_list.append(q_name_)
                q_name = q_name_
                
        for qName in q_name_list:
            best_score = -1e10
            best_tName = None
            best_tSize = None
            best_alignment = None 

            #print blast_result.q_name
            if (len(blast_result.q_name)==0): continue
            #sys.stdout.write("#################\n## query %s\n#################\n" %(qName))
            max_blast_score = None 

            # determining the maximum blast score
            q_names = copy(blast_result.q_name)
            scores = copy(blast_result.score)

            while (len(q_names) > 0) and (q_names[0] == qName):
                score = float(scores.pop(0))
                if max_blast_score is None:
                    max_blast_score = score
                else:
                    if score > max_blast_score:
                        max_blast_score = score
                q_names.pop(0)
                                
            #print 'max_blast_score:', max_blast_score
                        
            while (len(blast_result.q_name) > 0) and (blast_result.q_name[0] == qName):
                sys.stderr.write('est_entry: %s, remaining blast hits: %d\n' %(qName,len(blast_result.q_name)))

                score = float(blast_result.score.pop(0))
                qName_ = blast_result.q_name.pop(0)
                tName = blast_result.t_name.pop(0)
                qStrand_ = blast_result.q_strand.pop(0)
                tStrand = blast_result.t_strand.pop(0)
                qStart = blast_result.q_start.pop(0)
                qEnd = blast_result.q_end.pop(0)
                tStart = blast_result.t_start.pop(0)
                tEnd = blast_result.t_end.pop(0)

                if score < max_blast_score*parameters['min_relative_score']:
                    print 'ignoring alignment: ',score,max_blast_score*parameters['min_relative_score']
                    continue 
                                        
                assert(tStrand == '+')
                qSeq_ = est_fasta_dict[qName] 
                qSize = len(qSeq_) ;
                
                if qStrand_ == '+':
                    if qStart > 1:
                        tStart -= int(self.model.intron_len_max)
                    else:
                        tStart -= 5 # fudge 
                    if (qSize - qEnd) > 0:
                        tEnd += int(self.model.intron_len_max)
                    else:
                        tEnd += 5 # fudge 
                else:
                    if qStart > 1:
                        tEnd += int(self.model.intron_len_max)
                    else:
                        tEnd += 5 # fudge 
                    if (qSize - qEnd) > 0:
                        tStart -= int(self.model.intron_len_max)
                    else:
                        tStart -= 5 # fudge 
                
                for tStrand in '+-':
                    qStrand = qStrand_
                    if tStrand == '-':
                        if qStrand == '+': qStrand = '-'
                        else: qStrand = '+'
                                
                    if tStrand == '-' and parameters['respectTargetOrientation']: continue
                    if qStrand == '-' and parameters['respectQueryOrientation']: continue
                                
                    qSeq = qSeq_
                    if qStrand == '-':
                        qSeq = palma.palma_utils.reverse_complement(qSeq) 
                
                    tStart -= 1 
                    if tStart<0: tStart=0
                    if tEnd>len(dna_fasta_dict[tName]): tEnd = len(dna_fasta_dict[tName])

                    #print 'TRANSCRIPT (%s): chrom: %s, [%d:%d]' %(tStrand, tName, tStart, tEnd)
                    #print 'QUERY (%s):       %s, [%d:%d]' %(qStrand, qName, qStart, qEnd)
					
                    tSeq = dna_fasta_dict[tName][tStart:tEnd]
                    tSize = len(tSeq) 
                    if tStrand == '-':
                        tSeq = palma.palma_utils.reverse_complement(tSeq) 
                        
                    if tStrand == '+':
                        acc_support_values = self.get_splicescores(splicescores, tName + '_acceptor_splicescores', tStart, tEnd)
                        don_support_values = self.get_splicescores(splicescores, tName + '_donor_splicescores', tStart, tEnd)
                    else:
                        assert(tSeq == dna_fasta_dict[tName + '_reverse'][len(dna_fasta_dict[tName])-tEnd:len(dna_fasta_dict[tName])-tStart])
                        acc_support_values = self.get_splicescores(splicescores, tName + '_reverse_acceptor_splicescores', \
                                                                   len(dna_fasta_dict[tName])-tEnd, len(dna_fasta_dict[tName])-tStart)
                        don_support_values = self.get_splicescores(splicescores, tName + '_reverse_donor_splicescores', \
                                                                   len(dna_fasta_dict[tName])-tEnd, len(dna_fasta_dict[tName])-tStart)

                    #print len(tSeq), len(acc_support_values), len(don_support_values)
                    #<---------------------------------------------------------------------->
                    # get alignment information for nr_path many alignments
                    #<---------------------------------------------------------------------->       
                    align_info = [] ;
                    try:
                        reduce_region = (tSize*qSize*8 > parameters['reduce_threshold'])
                        align_info = palma.do_alignment(tSeq, tSize, qSeq, qSize, reduce_region,\
                                                        self.model, tStrand, translation, comparison_char,\
                                                        don_support_values, acc_support_values)
                
                    except MemoryError:
                        sys.stderr.write("WARNING: Exception caught. Probably not enough memory for creating the alignment matrices\n")

                    if len(align_info)>0:
                        align_info = align_info[0] # best path only
                        #print align_info.score
                        #print qSize, align_info.exon_length
                        if (align_info.score > best_score):# or (not parameters['bestAlignmentOnly']):
                            best_score = align_info.score
                            if tStrand == '+':
                                align_info.tStart += tStart 
                                align_info.tEnd += tStart 
                                align_info.tStarts = map(lambda x: x + tStart, align_info.tStarts)
                                align_info.tEnds = map(lambda x: x + tStart, align_info.tEnds)
                                align_info.tIDX = map(lambda x: x + tStart, align_info.tIDX)
                            else:
                                tStart=align_info.tStart
                                align_info.tStart = tEnd - align_info.tEnd + 1
                                align_info.tEnd = tEnd - tStart + 1
                                tStarts = align_info.tStarts
                                align_info.tStarts = map(lambda x: tEnd - x + 1, align_info.tEnds)
                                align_info.tEnds = map(lambda x: tEnd - x + 1, tStarts)
                                align_info.tIDX = map(lambda x: tEnd - x + 1, align_info.tIDX)

                            best_alignment = align_info
                            best_tStrand = tStrand
                            best_qStrand = qStrand
                            best_tName = tName
                            best_tSize = len(dna_fasta_dict[tName])      

                if (best_alignment is not None) and (not parameters['bestAlignmentOnly']):
                    palma.output_formating.print_results(best_alignment, out_file, bed_file, best_tName, \
                                                         best_tSize, qName, qSize, best_qStrand, \
                                                         best_tStrand, parameters)
                    best_score = -1e10 # reset 

            #sys.stderr.write('Done with alignment, printing...\n\n')
            if (best_alignment is not None) and (parameters['bestAlignmentOnly']):
                palma.output_formating.print_results(best_alignment, out_file, bed_file, best_tName, \
                                                     best_tSize, qName, qSize, best_qStrand, \
                                                     best_tStrand, parameters)


def parse_options():
    # default: action="store", type="string"
    parser = OptionParser(version="%prog version 0.3.6 (1 May 2007)\n(w) Uta Schulze, Bettina Hepp, Gunnar Raetsch, Soeren Sonnenburg, Cheng Soon Ong",
                          usage = "usage: %prog [options] <target-file> <query-file> <output-file>",
                          description="PALMA performs a local alignment of two nucleic sequences using organism specific splice site predictions and intron length information. The 'target' and 'query' files have to be in FASTA format. An 'U' is treated as 'T' and other characters than 'acgtnuACGTNU' are not allowed. The 'output' argument contains the file name with the alignment output. ")
        
    # we compute only the best path -> nr_paths=1           
    # parser.add_option("-n",
    #         action="store", type="int", dest="nr_paths",
    #         default=1, metavar="N",
    #         help="compute the N-best alignments (default: %default)")
        
    parser.add_option("--pslx", action="store_true", dest="print_sequences", default=False, 
                      help="Changes output file format from psl to pslx.\n\
                      psl  - Default. Tab separated format, no sequence\n\
                      pslx - Tab separated format with sequence")
    
    parser.add_option("--psl2", action="store_true", dest="print_psl2", default=False,
                      help="Use modified version of BLAT output instead,\n\
                      which gives blocks on both query and target sequence\n\
                      as well as PALMA's alignment score (default: False)")
        
    parser.add_option("--bed",
                      action="store", dest="bed_filename",
                      default="", metavar="string",
                      help="Optional output file (.bed format)")

    parser.add_option("--noHead",
                      action="store_true", dest="print_header", default=False,
                      help="suppress header (so it's just a tab-separated file) (default: False)")
        
    parser.add_option("--organism",
                      action="store",  dest="organism",
                      default="elegans", metavar="string", choices=["elegans","elegans_uncut","human"],
                      help="Defines which organism-dependent parameters are used.\n\
                      Available: elegans (default) and human.")
        
    parser.add_option("--noGC",
                      action="store_false", dest="use_gc", default=True,
                      help="Do not use GC donor splice sites, i.e. only GT sites")
        
    #parser.add_option("--pairedPrediction",
    #                                 action="store_true", dest="predict_paired_only", default=False,
    #                                 help="Assumes that the number of query sequences is equal to the number of \n\
    #                                 target sequences. Alignment is only performed on corresponding pairs of sequences.\n\
    #                                 Otherwise, all pairs of sequences are considered.")

    parser.add_option("--noSSscores",
                      action="store_false", dest="use_splicesites", default=True, 
                      help="Do not use splice site scores for the alignment")

    parser.add_option("--saveSSscores",
                      action="store_true", dest="save_splicescores", default=False, 
                      help="Saves the computed splice scores for a fasta file")

    parser.add_option("--noIntronLength",
                      action="store_false", dest="use_intronlength", default=True, 
                      help="Do not use intron length scores for the alignment")

    parser.add_option("--respectTargetOrientation",
                      action="store_true", dest="respectTargetOrientation", default=False,
                      help="If true, then the negative strand of the target (DNA) is not considered  (default: False)")

    parser.add_option("--respectQueryOrientation",
                      action="store_true", dest="respectQueryOrientation", default=False, 
                      help="If true, then the negative strand of the query (mRNA) is not considered (default: False)")
        
    parser.add_option("--useBlastn",
                      action="store_true", dest="use_blastn",
                      default=False, 
                      help="Use WU-Blastn to narrow down the search space")

    parser.add_option("--blastDb",
                      action="store", type="string", dest="blast_dbfile",
                      default="", metavar="string",
                      help="Specify the blast index file matching the target sequences\n\
                      (without the .xnd, .xns or .xnt ending)")

    parser.add_option("--maxEValue",
                      action="store", type="float", dest="max_e_value", default=1e-3,
                      help="Largest E-value accepted from blastn (default: 1e-3)")
    
    parser.add_option("--minRelativeScore",
                      action="store", type="float", dest="min_relative_score", default=0.75,
                      help="To filter suboptimal blastn hits, we compute the fraction \n\
                      of the score of matches relative to best blastn match.  \n\
                      Only if the fraction is above the specified threshold the match \n\
                      is considered (default: %default)")

    parser.add_option("--bestAlignmentOnly",
                      action="store_true", dest="bestAlignmentOnly",
                      default=False, 
                      help="Only output the best alignment")

    parser.add_option("--xdformat",
                      action="store", type="string", dest="xdformat",
                      default="xdformat", metavar="string",
                      help="Filename of the WU-blast xdformat command")
    
    parser.add_option("--blastn",
                      action="store", type="string", dest="blastn",
                      default="blastn", metavar="string",
                      help="Filename of the WU-blast blastn command")
    
    parser.add_option("--useSge",
                      action="store_true", dest="use_sge", default=False, 
                      help="Uses Sge for parallelization. Option not yet available.")


    #parser.add_option("-n",
    #                                 action="store", type="int", dest="nr_paths",
    #                                 default=1, metavar="N",
    #                                 help="compute the N-best alignments (default: %default)")


    #option for training type (on microexons(1) or real exons(2))

    (options, args) = parser.parse_args()
        
    filenames = {}
    parameters = {}
        
    #arguments
    if len(args) != 3:
        parser.error("incorrect number of arguments")

    filenames['dna'] = args[0]
    filenames['est'] = args[1]
    filenames['out'] = args[2]
    filenames['blast_out'] = args[2] + '_blast'
    filenames['bed'] = options.bed_filename
                
    parameters['print_header'] = not options.print_header
    parameters['print_psl2'] = options.print_psl2
    parameters['print_sequences'] = options.print_sequences
    parameters['blast_dbfile'] = options.blast_dbfile
    parameters['max_e_value'] = options.max_e_value
    parameters['min_relative_score'] = options.min_relative_score
    parameters['predict_paired_only'] = False #options.predict_paired_only
    parameters['bestAlignmentOnly'] = options.bestAlignmentOnly
    
    parameters['use_gc'] = options.use_gc
    parameters['use_splicesites'] = options.use_splicesites
    parameters['use_intronlength'] = options.use_intronlength
    parameters['save_splicescores'] = options.save_splicescores 
    
    parameters['use_blastn'] = options.use_blastn
    parameters['xdformat'] = options.xdformat
    parameters['blastn'] = options.blastn
    parameters['topcomboN'] = 1                           #see information about wublast above

    # extensions
    parameters['use_sge'] = options.use_sge
    
    #organism and dependet options
    parameters['organism'] =          options.organism
    parameters['max_gene_length']    = organism_constants[options.organism]['max_gene_length']
    parameters['genome_version']    = organism_constants[options.organism]['version']
    
    parameters['respectTargetOrientation'] = options.respectTargetOrientation
    parameters['respectQueryOrientation'] = options.respectQueryOrientation
    
    parameters['reduce_threshold'] = 1024**3

    #if (not parameters['use_blastn']) and (parameters['respectQueryOrientation']):
        #parameters['respectTargetOrientation'] = True
        #parameters['respectQueryOrientation'] = True
                
    filenames['param'] =    options.organism + '_' + parameters['genome_version'] + '_ss=' \
                         + `int(parameters['use_splicesites'])` + '_il=' \
                         + `int(parameters['use_intronlength'])` + param_filename_ending
        
    #<------------------------------------------------------------------------------>
    #filenames['dir'] = './palma'
    #filenames['data_dir'] = './palma/data/'
    filenames['dir'] = palma.__path__[0]
    filenames['data_dir'] = palma.__path__[0] + '/data/'
        
    #<------------------------------------------------------------------------------>
    # ss type (score, pred or none)
    #<------------------------------------------------------------------------------>
    if options.use_splicesites:
        try:
            #sys.stdout.write("Trying to import shogun..\n")
            from shogun.Classifier import SVM
            from shogun.Features import CharFeatures,DNA
            from shogun.Kernel import WeightedDegreeCharKernel
        except:
            sys.stderr.write("'import shogun' failed. (See README and shogun website: http://www.shogun-toolbox.org)\n")
            sys.exit(2)
    #<------------------------------------------------------------------------------>

    return (parameters, filenames)

# main
if __name__ == '__main__':
    (parameters, filenames) = parse_options()

    #<------------------------------------------------------------------------------>
    # open output-file and print info line in output-file
    #<------------------------------------------------------------------------------>
    try:
        out_file = open(filenames['out'],'w+')
    except IOError:
        sys.stderr.write("Cannot open %s to write\n" %out_filename)
        sys.exit(2)

    if len(filenames['bed']) > 0:
        try:
            bed_file = open(filenames['bed'],'w')
        except IOError:
            sys.stderr.write("Cannot open %s to write\n" %filenames['bed'])
            sys.exit(2)
    else:
        bed_file = None
    #<------------------------------------------------------------------------------>

    p=palma_align()
    # load the model
    p.load_model(filenames, parameters)

    t0 = time.time()
    dna_fasta_dict = palma.genomic.read_fasta(file(filenames['dna']), {})
    # DNA sequences
    sys.stderr.write('found DNA fasta file with ' + `len(dna_fasta_dict)` + ' sequence(s)\n')
    if not parameters['respectTargetOrientation'] or parameters['save_splicescores']:
        keys = dna_fasta_dict.keys()
        for s in keys:
			seq = dna_fasta_dict[s]
			dna_fasta_dict[s + '_reverse'] = palma.genomic.reverse_complement(seq)
        #print dna_fasta_dict.keys()

    # EST sequences
    est_fasta_dict = palma.genomic.read_fasta(file(filenames['est']), {})

    t1 = time.time()
    # predict splice sites for all dna sequences
    splicescores = p.predict_splicescores(filenames, parameters, dna_fasta_dict)

    if parameters['respectTargetOrientation']:
        keys = dna_fasta_dict.keys()
        for s in keys:
            if s.endswith('_reverse'):
                del dna_fasta_dict[s] ;
        #print dna_fasta_dict.keys()
		
    t2 = time.time()
    if parameters['print_header']:
        if parameters['print_psl2']:
            palma.output_formating.print_header(out_file)
        else:
            palma.output_formating.print_blat_header(out_file)

    if parameters['use_blastn']:
        p.xdformat(filenames, parameters) 
        blast_result = p.blastn(filenames, parameters)
        t3 = time.time()
        p.predict_file_blast(filenames, parameters, out_file, bed_file, dna_fasta_dict, est_fasta_dict, splicescores, blast_result)
    else:
        t3 = time.time()
        if parameters['predict_paired_only']:
            p.predict_file_paired(filenames, parameters, out_file, bed_file, dna_fasta_dict, est_fasta_dict, splicescores)
        else:
            p.predict_file_all(filenames, parameters, out_file, bed_file, dna_fasta_dict, est_fasta_dict, splicescores)
    t4 = time.time()

    print 'Timing '
    print 'Reading fasta files:\t\t\t', t1-t0, 'seconds'
    if parameters['use_splicesites']:
        print 'Initializing splice site scores:\t', t2-t1, 'seconds'
    if parameters['use_blastn']:
        print 'BLAST:\t\t\t\t\t', t3-t2, 'seconds'
    print 'PALMA alignment:\t\t\t', t4-t3, 'seconds'

    #<------------------------------------------------------------------------------>
    out_file.close()
    if bed_file is not None: bed_file.close()
        
    sys.exit(1)
    #<------------------------------------------------------------------------------>


