/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Written (W) 2004-2006 Uta Schulze
 * Written (W) 2004-2006 Bettina Hepp
 * Written (W) 2004-2006 Gunnar Raetsch
 * Copyright (C) 2004-2006 Fraunhofer Institute FIRST and Max-Planck-Society
 */

/*********************************************************************************************/
// Version of fill_matrix with fills several alignment matrices (best, second_best ...)
// nr_paths det. number of these matrices
//
// est_len: Number of columns in alignment matrix: length(EST) +1 (for the gap) (variable: i)
// dna_len: Number of lines in alignment matrix: length(DNA) +1 (for the gap) (variable: j)
//
// !!! IMPORTANT !!! The pictures and equations in the paper/poster etc. describe the algorithm
// for an alignment matrix where DNA and EST changed places !!! IMPORTANT !!!
/*********************************************************************************************/

/*
  matchmatrix: columnwise
    - A C G T N (dna)
  - x x x x x x
  A x x x x x x	
  C x x x x x x
  G x x x x x x
  T x x x x x x
  N x x x x x x
(est)
*/

#include <math.h>
#include "myalign_1best.h"
#include "Mathmatics.h"
#include <assert.h>
#include <algorithm> //Eigentlich in Header

using namespace std;

double bound_matrix(int est_len, int dna_len, char* est, char* dna, struct penalty h, double* matchmatrix, double* donor, double* acceptor, int nr_donor_sites, int* donor_sites, int* max_score_position)
{
  
  double* prev_column = new double[dna_len];
  if (!prev_column) {throw MyAlignException() ;}
  
  double* act_column = new double[dna_len];
  if (!act_column) {throw MyAlignException() ;}
  
  for (int j=1; j<dna_len; j++) {
	  act_column[j] = log(0.0) ; // fills column with -INF
  }
  
  
  bool* isSplice = new bool[dna_len] ;
  if (!isSplice) {throw MyAlignException() ;}
  memset((bool*)isSplice, 0, (dna_len)*sizeof(bool)); // fills column with -INF
  
  /*********************************************************************************************/
  // variables
  /*********************************************************************************************/
  const int mlen = 6; // length of matchmatrix
  int estChar, dnaChar; // 0:'-', 1:'a', 2:'c', 3:'g', 4:'t', 5:'n'
  double prevValue ;
  double tempValue;
  int last_supp_point = (int)h.limits[h.len-1] ; //last (len) supporting point of h
  //fprintf(stdout,"last supp_point in h: %d\n", last_supp_point) ;
  
  int num_elem ; // counter variable
  
  const int max_num_elem = 4+last_supp_point ; // upper bound on num_elem
  double* scores = new double[max_num_elem] ; //are sorted later
  if (!scores) {throw MyAlignException() ;}
  struct ArrayElem *array=new ArrayElem[max_num_elem] ;
  if (!array) {throw MyAlignException() ;}
  //fprintf(stdout, "%d\n", max_num_elem*sizeof(ArrayElem)) ;
  
  int max_ss_element ; //position on dna with max ss score for introns of length greater than last_supp_point
  int splice_pos ; //splice position get from donor_sites
  int start_splice_pos ; //where to start looking in donor_sites
  
  double max_al_score = log(0.0) ; //maximal alignment score so far
  max_score_position[0] = 0 ; // pos i 
  max_score_position[1] = 0 ; // pos j 
  
  
  /*********************************************************************************************/	
  // Column Zero
  /*********************************************************************************************/
  memset((double*)prev_column, 0, (dna_len)*sizeof(double)); // fills column with -INF
  
  
  /*********************************************************************************************/	
  // Line Zero 
  /*********************************************************************************************/
  act_column[0] = 0 ;
  
  
  /* columnwise */
  for (int i=1; i<est_len; i++) {/* i.column */ /* est_len = len(est)+1 */
	  
	  //fprintf(stdout,"%d\t",i) ;
	  start_splice_pos = 0 ; //where to start looking in donor_sites
	  max_ss_element   = 0 ; //initializing
	  
	  for (int j=1; j<dna_len; j++) //j.line
	  {
		  dnaChar = check_char(dna[j-1]) ; // -1 because of of the gaps in the 0.line/column of the al-matrix
		  estChar = check_char(est[i-1]) ;
		  
		  
		  //if introns of length greater than last_supp_point possible
		  if (j > last_supp_point) {
			  //storing highest ss_score (the position) and NOT -INF
			  int new_ss_pos = j-last_supp_point ; //new possible donor_site
			  if (isfinite(donor[new_ss_pos-1])) { //splice site!
				  if (max_ss_element != 0) { 
					  double new_intron_score = donor[new_ss_pos-1] + act_column[new_ss_pos-1] ; //donor + prev_value
					  assert(isfinite(donor[max_ss_element-1])) ;
					  double act_intron_score = donor[max_ss_element-1]+ act_column[max_ss_element-1] ; //donor + prev_value
					  if (new_intron_score >= act_intron_score) {
						  max_ss_element = new_ss_pos ; //replacing
					  }
				  }
				  else {
					  max_ss_element = j-last_supp_point ;
				  }
			  }
			  //start_pos in donor_sites so that only introns shorter then "last_supp_point" are seen
			  if ((start_splice_pos < nr_donor_sites) && donor_sites[start_splice_pos] == j-last_supp_point) {
				  start_splice_pos++ ;
			  }
		  }
		  
		  
		  /* Computing all possible scores */
		  num_elem = 0 ; /* number of different scores/possibilities so far */
		  
		  /***************************************/
		  // 0. Zero (local alignment version)
		  /***************************************/
		  assert(num_elem<max_num_elem) ;
		  array[num_elem].prev_i = 0; 
		  array[num_elem].prev_j = 0; 
		  scores[num_elem] = 0 ;
		  num_elem++ ;
		  
		  
		  /***************************************/
		  // 1. gap on EST (j-1,i) -> (j,i) 
		  /***************************************/
		  prevValue = act_column[j-1] ; /* (j-1,i) -> (j,i) */
		  if (isfinite(prevValue)) {
			  tempValue = prevValue +(matchmatrix[mlen* dnaChar]); /* score(gap,DNA) */
			  if (isfinite(tempValue)) {
				  assert(num_elem<max_num_elem) ;
				  array[num_elem].prev_i = i; 
				  array[num_elem].prev_j = j-1; /* predecessor */
				  scores[num_elem] = -tempValue ;
				  num_elem++ ;
			  }
		  }
		  
		  
		  /***************************************/
		  // 2. match/mismatch (j-1,i-1) -> (j,i)
		  /***************************************/
		  prevValue = prev_column[j-1] ; /* (j-1,i-1) -> (j,i) */
		  if (isfinite(prevValue)) {
			  tempValue = prevValue +(matchmatrix[mlen* dnaChar +estChar]);//DNA mit EST
			  if (isfinite(tempValue)) {
				  assert(num_elem<max_num_elem) ;
				  array[num_elem].prev_i = i-1; /* predecessor */
				  array[num_elem].prev_j = j-1; /* predecessor */
				  scores[num_elem] = -tempValue ;
				  num_elem++ ;
			  }
		  }
		  
		  
		  /***************************************/
		  // 3. gap on DNA (j,i-1) -> (j,i)
		  /***************************************/
		  prevValue = prev_column[j] ;
		  if (isfinite(prevValue)) {
			  tempValue = prevValue +(matchmatrix[estChar]); /* score(EST,gap) */
			  if (isfinite(tempValue)) {
				  assert(num_elem<max_num_elem) ;
				  array[num_elem].prev_i = i-1; /* predecessor */
				  array[num_elem].prev_j = j;
				  scores[num_elem]=-tempValue ;
				  num_elem++ ;
			  }	
		  }
		  
		  
		  /***************************************/
		  // 4. all possible splice sites/introns
		  /***************************************/
		  // acceptor[j-1]: g of ag (-1 because cpp starts counting at 0)
		  // donor[splice_pos-1]: g of gt/gc (-1 because cpp starts counting at 0)
		  // donor_sites: the g's of the gt/gc
		  
		  if (isfinite(acceptor[j-1])) {
			  
			  //the nr_paths many introns of length greater than last_supp_point
			  if (max_ss_element>0) { //the g of the gt....
				  assert((j-max_ss_element+1) > last_supp_point) ;
				  
				  prevValue = act_column[max_ss_element-1] ; //one before the _gt....
				  if (isfinite(prevValue) && (!isSplice[max_ss_element-1])) {
					  int dist = j -(max_ss_element-1); //_gt...ag: dist from g to g
					  assert(dist>last_supp_point) ;
					  tempValue = penalty_lookup(h,dist) +donor[max_ss_element-1] +acceptor[j-1] +prevValue;
					  if (isfinite(tempValue)) {
						  assert(num_elem<max_num_elem) ;
						  array[num_elem].prev_i = i; 
						  array[num_elem].prev_j = max_ss_element-1; // predecessor
						  scores[num_elem]=-tempValue ;
						  num_elem++ ;
					  }
					  else if (isfinite(penalty_lookup(h,dist))) {
						  fprintf(stdout, "Something wrong with donor_sites(1000er) (intron_length: %d)\n", dist) ;
						  sleep(5000) ;
					  }
				  }
			  }
			  
			  //all introns of length <= last_supp_point
			  for (int k=start_splice_pos; k < nr_donor_sites; k++) {
				  //for (int k=0; k < nr_donor_sites; k++) //also intron of length greater than 1000
				  splice_pos = donor_sites[k] ;
				  if (splice_pos > j-4) { break ;}
				  prevValue = act_column[splice_pos-1] ; //one before the _gt
				  
				  if (isfinite(prevValue) && (!isSplice[splice_pos-1])) {
					  int dist = j -(splice_pos-1); //_gt...ag: dist from g to g
					  assert(dist<=last_supp_point) ;
					  tempValue = penalty_lookup(h,dist) +donor[splice_pos-1] +acceptor[j-1] +prevValue;
					  if (isfinite(tempValue)) {
						  assert(num_elem<max_num_elem) ;
						  array[num_elem].prev_i = i; 
						  array[num_elem].prev_j = splice_pos-1; /* predecessor */
						  scores[num_elem]=-tempValue ;
						  num_elem++ ;
					  }
					  else if (isfinite(penalty_lookup(h,dist))) {
						  fprintf(stdout, "Something wrong with donor_sites (intron_length: %d)\n", dist) ;
						  sleep(5000) ;
					  }
				  }
			  }
		  }
		  
		  CMath::nmin<struct ArrayElem>(scores, array, num_elem, 1); //highest one
		  
		  /* the highest score is written in the matrix */
		  if (num_elem!=0) {
			  act_column[j] = -scores[0] ;
			  if (array[0].prev_j < j-1 && -scores[0]!=0) {
				  isSplice[j] = 1 ;
				  //if (i<4) { fprintf(stdout,"bound_matrix: splicesite:[%d,%d]\n",i,j) ;}
			  }
			  if (-scores[0]>max_al_score) { //storing the position of the highest alignment score so far
				  max_al_score = -scores[0] ;
				  //fprintf(stdout,"bound_matrix[%d,%d]: %f\n", i, j, max_al_score) ;
				  max_score_position[0] = i; // pos i
				  max_score_position[1] = j; // pos j
			  }
		  }
	  } //end of j
	  
	  {
		  double * tmp = prev_column ;
		  prev_column = act_column ;
		  act_column = tmp ;
	  }
	  
	  act_column[0] = 0 ; //0. line
	  for (int jj=1; jj<dna_len; jj++) {
		  act_column[jj] = log(0.0) ; // fills column with -INF
	  }
	  memset((bool*)isSplice, 0, (dna_len)*sizeof(bool)); // fills column with -INF
  } // end of i
  //fprintf(stdout,"\n") ;

  delete[] prev_column ;
  delete[] act_column ;

  //free memory
  delete[] array;
  delete[] scores;
  delete[] isSplice ;
  
  return max_al_score;
}
